;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "plantuml-mode" "1.9.0"
  "Major mode for PlantUML."
  '((dash    "2.0.0")
    (emacs   "25.1")
    (deflate "0.0.3"))
  :url "https://github.com/skuro/plantuml-mode"
  :commit "7a93f7ce96cff5bd48ca4cde70d3bbf789839613"
  :revdesc "v1.9.0-0-g7a93f7ce96cf"
  :keywords '("files" "text" "processes" "tools"))
